# -*- coding: utf-8 -*-

r'''This module contains a class for handling batched datasets.

In many optimization tasks, parameters must be updated by optimizing them with
respect to estimates of a loss function. The loss function for many problems is
estimated using a set of data that we have measured.
'''

import climate
import collections
import numpy.random as rng

logging = climate.get_logger(__name__)


class Dataset:
    '''This class handles batching and shuffling a dataset.

    In ``downhill``, losses are optimized using sets of data collected from the
    problem that generated the loss.

    During optimization, data are grouped into "mini-batches"---that is, chunks
    that are larger than 1 sample and smaller than the entire set of samples;
    typically the size of a mini-batch is between 10 and 100, but the specific
    setting can be varied depending on your model, hardware, dataset, and so
    forth. These mini-batches must be presented to the optimization algorithm in
    pseudo-random order to match the underlying stochasticity assumptions of
    many optimization algorithms. This class handles the process of grouping
    data into mini-batches as well as iterating and shuffling these mini-batches
    dynamically as the dataset is consumed by the optimization algorithm.

    For many tasks, a dataset is obtained as a large block of sample data, which
    in Python is normally assembled as a ``numpy`` ndarray. To use this class on
    such a dataset, just pass in a list or tuple containing ``numpy`` arrays;
    the number of these arrays must match the number of inputs that your loss
    computation requires.

    There are some cases when a suitable set of training data would be
    prohibitively expensive to assemble in memory as a single ``numpy`` array.
    To handle these cases, this class can also handle a dataset that is provided
    via a Python callable. For more information on using callables to provide
    data to your model, see :ref:`quickstart-using-callables`.

    Parameters
    ----------
    inputs : ndarray, tuple, list, or callable
        One or more sets of data.

        If this parameter is callable, then mini-batches will be obtained by
        calling the callable with no arguments; the callable is expected to
        return a tuple of ndarrays that will be suitable for optimizing the
        loss at hand.

        If this parameter is a list or tuple, it must contain ndarrays (or
        something similar with a ``shape`` attribute, like a pandas DataFrame).
        These are assumed to contain data for computing the loss; the length of
        this tuple or list should match the number of inputs required by the
        loss computation. If multiple ndarrays are provided, their lengths along
        the axis given by the ``axis`` parameter (defaults to 0) must match.

    name : str, optional
        A string that is used to describe this dataset. Usually something like
        'test' or 'train'.

    batch_size : int, optional
        The size of the mini-batches to create from the data sequences. Defaults
        to 32.

    iteration_size : int, optional
        The number of batches to yield for each call to iterate(). Defaults to
        the length of the data divided by batch_size. If the dataset is a
        callable, then the number is len(callable). If callable has no length,
        then the number is set to 100.

    axis : int, optional
        The axis along which to split the data arrays, if the first parameter is
        given as one or more ndarrays. If not provided, defaults to 0.
    '''

    _count = 0

    def __init__(self, inputs, name=None, batch_size=32, iteration_size=None, axis=0):
        '''Create a minibatch dataset from data arrays or a callable.'''
        self.name = name or 'dataset{}'.format(Dataset._count)
        Dataset._count += 1
        self.batch_size = batch_size
        self.iteration_size = iteration_size

        self._batches = None
        self._callable = None

        if isinstance(inputs, collections.Callable):
            self._init_callable(inputs)
        else:
            self._init_arrays(inputs, axis)

    def _init_callable(self, inputs):
        self._callable = inputs
        if not self.iteration_size:
            try:
                self.iteration_size = len(inputs)
            except TypeError:  # has no len
                self.iteration_size = 100
        logging.info('%s: %d mini-batches from callable',
                     self.name, self.iteration_size)

    def _init_arrays(self, inputs, axis=0):
        if not isinstance(inputs, (tuple, list)):
            inputs = (inputs, )

        L = inputs[0].shape[axis]
        assert all(L == x.shape[axis] for x in inputs), \
            'shapes do not match along axis {}: {}'.format(
                axis, '; '.join(str(x.shape) for x in inputs))

        self._index = 0
        self._batches = []
        for i in range(0, L, self.batch_size):
            batch = []
            for x in inputs:
                slices = [slice(None) for _ in x.shape]
                slices[axis] = slice(i, i + self.batch_size)
                b = x[tuple(slices)]
                if b.shape[axis] != self.batch_size:
                    break
                batch.append(b)
            else:
                self._batches.append(batch)

        self.shuffle()

        if not self.iteration_size:
            self.iteration_size = len(self._batches)

        logging.info('%s: %d of %d mini-batches of %s',
                     self.name,
                     self.iteration_size,
                     len(self._batches),
                     '; '.join(str(x.shape) for x in self._batches[0]))

    def __iter__(self):
        return self.iterate(True)

    def shuffle(self):
        '''Shuffle the batches in the dataset.

        If this dataset was constructed using a callable, this method has no
        effect.
        '''
        if self._batches is not None:
            rng.shuffle(self._batches)

    def iterate(self, shuffle=True):
        '''Iterate over batches in the dataset.

        This method generates ``iteration_size`` batches from the dataset and
        then returns.

        Parameters
        ----------
        shuffle : bool, optional
            Shuffle the batches in this dataset if the iteration reaches the end
            of the batch list. Defaults to True.

        Returns
        -------
        batches : sequence of data batches
            A sequence of batches---often from a training, validation, or test
            dataset.
        '''
        for _ in range(self.iteration_size):
            if self._callable is not None:
                yield self._callable()
            else:
                yield self._next_batch(shuffle)

    def _next_batch(self, shuffle=True):
        value = self._batches[self._index]
        self._index += 1
        if self._index >= len(self._batches):
            if shuffle:
                self.shuffle()
            self._index = 0
        return value



# -*- coding: utf-8 -*-

'''A module of utility functions and other goodies.'''

import numpy as np
import theano
import theano.tensor as TT


class Registrar(type):
    '''A metaclass that builds a registry of its subclasses.'''

    def __init__(cls, name, bases, dct):
        if not hasattr(cls, '_registry'):
            cls._registry = {}
        else:
            cls._registry[name.lower()] = cls
        super(Registrar, cls).__init__(name, bases, dct)

    def build(cls, key, *args, **kwargs):
        return cls._registry[key.lower()](*args, **kwargs)

    def get_class(cls, key):
        return cls._registry[key.lower()]

    def is_registered(cls, key):
        return key.lower() in cls._registry


def shared_like(param, suffix, init=0):
    '''Create a Theano shared variable like an existing parameter.

    Parameters
    ----------
    param : Theano variable
        Theano variable to use for shape information.
    suffix : str
        Suffix to append to the parameter's name for the new variable.
    init : float or ndarray, optional
        Initial value of the shared variable. Defaults to 0.

    Returns
    -------
    shared : Theano shared variable
        A new shared variable with the same shape and data type as ``param``.
    '''
    return theano.shared(np.zeros_like(param.get_value()) + init,
                         name='{}_{}'.format(param.name, suffix))


def as_float(x):
    '''Cast a floating point value to a Theano ``floatX`` symbol.

    Parameters
    ----------
    x : float, ndarray, or Theano expression
        Some quantity to cast to floating point.

    Returns
    -------
    x : Theano expression
        A symbolic variable cast as a ``floatX`` value.
    '''
    return TT.cast(x, theano.config.floatX)


def as_float32(x):
    '''Cast a floating point value to a 32-bit numpy float.

    Parameters
    ----------
    x : float
        A value to cast to floating point.

    Returns
    -------
    x : np.float32
        The given value as a 32-bit float.
    '''
    return np.float32(x)



'''This module defines a base class for optimization techniques.'''

import climate
import collections
import numpy as np
import theano
import theano.tensor as TT

from . import util

logging = climate.get_logger(__name__)


def build(algo, loss, params, inputs, updates=(), monitors=(),
          monitor_gradients=False):
    '''Construct an optimizer by name.

    Parameters
    ----------
    algo : str
        The name of the optimization algorithm to build.
    loss : Theano expression
        Loss function to minimize. This must be a scalar-valued expression.
    params : list of Theano variables
        Symbolic variables to adjust to minimize the loss.
    inputs : list of Theano variables
        Symbolic variables required to compute the loss.
    updates : list of update pairs, optional
        A list of pairs providing updates for the internal of the loss
        computation. Normally this is empty, but it can be provided if the loss,
        for example, requires an update to an internal random number generator.
    monitors : dict or sequence of (str, Theano expression) tuples, optional
        Additional values to monitor during optimization. These must be provided
        as either a sequence of (name, expression) tuples, or as a dictionary
        mapping string names to Theano expressions.
    monitor_gradients : bool, optional
        If True, add monitors to log the norms of the parameter gradients during
        optimization. Defaults to False.

    Returns
    -------
    optimizer : :class:`Optimizer`
        An optimizer instance.
    '''
    return Optimizer.build(algo, loss, params, inputs,
                           updates=updates, monitors=monitors,
                           monitor_gradients=monitor_gradients)


class Optimizer(util.Registrar(str('Base'), (), {})):
    '''An optimizer computes gradient updates to iteratively optimize a loss.

    Parameters
    ----------
    loss : Theano expression
        Loss function to minimize. This must be a scalar-valued expression.
    params : list of Theano variables
        Symbolic variables to adjust to minimize the loss.
    inputs : list of Theano variables
        Symbolic variables required to compute the loss.
    updates : list of update pairs, optional
        A list of pairs providing updates for the internal of the loss
        computation. Normally this is empty, but it can be provided if the loss,
        for example, requires an update to an internal random number generator.
    monitors : dict or sequence of (str, Theano expression) tuples, optional
        Additional values to monitor during optimization. These must be provided
        as either a sequence of (name, expression) tuples, or as a dictionary
        mapping string names to Theano expressions.
    monitor_gradients : bool, optional
        If True, add monitors to log the norms of the parameter gradients during
        optimization. Defaults to False.
    '''

    def __init__(self, loss, params, inputs, updates=(), monitors=(),
                 monitor_gradients=False):
        self._loss = loss
        self._params = params
        self._inputs = inputs
        self._updates = updates
        if hasattr(updates, 'items') and callable(updates.items):
            self._updates = updates.items()

        self._shapes = [p.get_value(borrow=True).shape for p in self._params]
        self._counts = [np.prod(s) for s in self._shapes]
        self._starts = np.cumsum([0] + self._counts)[:-1]
        self._dtype = self._params[0].get_value().dtype

        self._curr_iter = 0
        self._best_iter = 0
        self._best_loss = 1e100
        self._best_params = [p.get_value().copy() for p in self._params]

        if hasattr(monitors, 'items') and callable(monitors.items):
            monitors = monitors.items()
        self._monitor_exprs = [self._loss]
        self._monitor_names = ['loss']
        for name, monitor in monitors:
            self._monitor_names.append(name)
            self._monitor_exprs.append(monitor)
        if monitor_gradients:
            for p, g in zip(self._params, TT.grad(self._loss, self._params)):
                self._monitor_names.append('∂{}'.format(p.name))
                self._monitor_exprs.append((g * g).sum())

    def _compile(self):
        '''Compile the Theano functions for evaluating and updating our model.
        '''
        logging.info('compiling evaluation function')
        self.f_eval = theano.function(
            self._inputs, self._monitor_exprs, updates=self._updates)
        logging.info('compiling %s step function', self.__class__.__name__)
        updates = list(self._updates) + list(self._get_updates())
        self.f_step = theano.function(
            self._inputs, self._monitor_exprs, updates=updates)

    def _get_updates(self):
        '''Get parameter update expressions for performing optimization.

        Returns
        -------
        updates : sequence of (parameter, expression) tuples
            A sequence of parameter updates to be applied during optimization.
        '''
        for param, grad in self._differentiate():
            for var, expr in self._get_updates_for(param, grad):
                if self.momentum == 0 or var != param:
                    yield var, expr
                    continue
                delta = expr - param
                vel_tm1 = util.shared_like(param, 'vel')
                vel_t = self.momentum * vel_tm1 + delta
                yield vel_tm1, vel_t
                if self.nesterov:
                    # see http://arxiv.org/pdf/1212.0901v2.pdf (eq 7) and
                    # https://github.com/lisa-lab/pylearn2/pull/136#issuecomment-10381617
                    yield param, (param + self.momentum ** 2 * vel_tm1
                                  + (1 + self.momentum) * delta)
                else:
                    yield param, param + vel_t

    def _get_updates_for(self, param, grad):
        '''Generate some update pairs for the given model parameter.

        Returns
        -------
        updates : sequence of (parameter, expression) tuples
            A sequence of parameter updates to be applied during optimization.
        '''
        raise NotImplementedError

    def _differentiate(self, params=None):
        '''Return a sequence of gradients for our parameters.

        This method applies gradient norm clipping, so if a gradient has a norm
        that exceeds the threshold, it will be rescaled to fit within the norm
        threshold.

        Parameters
        ----------
        params : list of Theano variables, optional
            Return the gradient with respect to these parameters. Defaults to
            all parameters that the optimizer knows about.

        Returns
        -------
        pairs : sequence of (param, grad) tuples
            Generates a sequence of tuples representing each of the parameters
            requested and the corresponding Theano gradient expressions.
        '''
        if params is None:
            params = self._params
        for param, grad in zip(params, TT.grad(self._loss, params)):
            norm = TT.sqrt((grad * grad).sum())
            yield param, TT.clip(
                grad * TT.minimum(1, self.max_gradient_norm / norm),
                -self.max_gradient_clip, self.max_gradient_clip)

    def set_params(self, targets):
        '''Set the values of the parameters to the given target values.

        Parameters
        ----------
        targets : sequence of ndarray
            Arrays for setting the parameters of our model.
        '''
        for param, target in zip(self._params, targets):
            param.set_value(target)

    def _log(self, monitors, iteration, label='', suffix=''):
        '''Log the state of the optimizer through the logging system.

        Parameters
        ----------
        monitors : OrderedDict
            A dictionary of monitor names mapped to values. These names and
            values are what is being logged.
        iteration : int
            Optimization iteration that we are logging.
        label : str, optional
            A label for the name of the optimizer creating the log line.
            Defaults to the name of the current class.
        suffix : str, optional
            A suffix to add to the end of the log line, if any.
        '''
        label = label or self.__class__.__name__
        fields = (('{}={:.6f}').format(k, v) for k, v in monitors.items())
        logging.info('%s %i %s%s', label, iteration, ' '.join(fields), suffix)

    def evaluate(self, dataset):
        '''Evaluate the current model parameters on a dataset.

        Parameters
        ----------
        dataset : :class:`Dataset <downhill.dataset.Dataset>`
            A set of data to use for evaluating the model.

        Returns
        -------
        monitors : OrderedDict
            A dictionary mapping monitor names to values. Monitors are
            quantities of interest during optimization---for example, loss
            function, accuracy, or whatever the optimization task requires.
        '''
        values = [self.f_eval(*x) for x in dataset]
        monitors = zip(self._monitor_names, np.mean(values, axis=0))
        return collections.OrderedDict(monitors)

    def _test_patience(self, monitors):
        '''Test whether our patience with optimization has elapsed.

        Parameters
        ----------
        monitors : dict
            A dictionary mapping monitor names to values. The 'loss' key from
            this dictionary will be used to evaluate optimization progress.

        Returns
        -------
        elapsed : bool
            True iff our patience has elapsed and the model is no longer
            improving.
        '''
        self._curr_iter += 1
        marker = ''
        loss = monitors['loss']
        if self._best_loss - loss > self._best_loss * self.min_improvement:
            self._best_loss = loss
            self._best_iter = self._curr_iter
            self._best_params = [p.get_value().copy() for p in self._params]
            marker = ' *'
        self._log(monitors, self._curr_iter - 1, 'validation', marker)
        return self._curr_iter - self._best_iter > self.patience

    def _prepare(self, **kwargs):
        '''Set up properties for optimization.

        This method can be overridden by base classes to provide parameters that
        are specific to a particular optimization technique (e.g., setting up a
        learning rate value).
        '''
        pass

    def iteropt(self,
                train,
                valid=None,
                patience=5,
                validate_every=10,
                min_improvement=0,
                max_gradient_norm=1e10,
                max_gradient_clip=1e10,
                learning_rate=1e-4,
                momentum=0,
                nesterov=False,
                **kwargs):
        r'''Optimize a loss iteratively using a training and validation dataset.

        This method yields a series of monitor values to the caller. After every
        optimization epoch, a pair of monitor dictionaries is generated: one
        evaluated on the training dataset during the epoch, and another
        evaluated on the validation dataset at the most recent validation epoch.

        The validation monitors might not be updated during every optimization
        iteration; in this case, the most recent validation monitors will be
        yielded along with the training monitors.

        Parameters
        ----------
        train : sequence or :class:`Dataset <downhill.dataset.Dataset>`
            A set of training data for computing updates to model parameters.
        valid : sequence or :class:`Dataset <downhill.dataset.Dataset>`, optional
            A set of validation data for computing monitor values and
            determining when the loss has stopped improving. Defaults to the
            training data.
        patience : int, optional
            Number of validation "failures" that we are willing to tolerate
            before stopping the optimization process. A validation failure
            happens whenever the loss on the validation dataset decreases by
            less than ``min_improvement`` (relative) over the previous best
            validation loss. Defaults to 5.
        validate_every : int, optional
            Evaluate the loss on the validation dataset after making this many
            passes over the training data. Defaults to 10.
        min_improvement : float, optional
            Insist that the validation loss must improve by this relative amount
            before considering that the optimization has made progress. The
            optimization process halts when ``patience`` validations have failed
            to make this relative improvement. Defaults to 0; set to a larger
            value (e.g., 0.01 for 1% improvement) to halt the optimization
            process sooner.
        max_gradient_norm : float, optional
            Rescale each parameter's gradient so that it has at most this L2
            norm. Defaults to 1e10, i.e., very little rescaling.
        max_gradient_clip : float, optional
            Perform elementwise clipping on gradient values (this happens after
            rescaling). Defaults to 1e10, i.e., very little clipping.
        learning_rate : float, optional
            Many SGD-based optimization algorithms require a learning rate
            hyperparameter that scales the gradient step. Defaults to 1e-4.
        momentum : float, optional
            Apply momentum to the parameter updates for this optimizer, with the
            given strength. Typically this value ranges from 0 (no momentum) to
            :math:`1 - \epsilon` (large momentum). Defaults to 0.
        nesterov : bool, optional
            If True, and momentum is nonzero, apply Nesterov-style momentum to
            parameter updates for this optimizer. If False and momentum is
            nonzero, "regular" momentum is applied. See
            :class:`NAG <downhill.NAG>` for a description of Nesterov momentum.

        Returns
        -------
        train_monitors : dict
            A dictionary mapping monitor names to values, evaluated on the
            training dataset.
        valid_monitors : dict
            A dictionary containing monitor values evaluated on the validation
            dataset.
        '''
        self.patience = patience
        self.validate_every = validate_every
        self.min_improvement = min_improvement
        self.max_gradient_norm = util.as_float(max_gradient_norm)
        self.max_gradient_clip = util.as_float(max_gradient_clip)
        self.learning_rate = util.as_float(learning_rate)
        self.momentum = util.as_float(momentum)
        self.nesterov = nesterov
        logging.info('-- patience = %s', patience)
        logging.info('-- validate_every = %s', validate_every)
        logging.info('-- min_improvement = %s', min_improvement)
        logging.info('-- max_gradient_norm = %s', max_gradient_norm)
        logging.info('-- max_gradient_clip = %s', max_gradient_clip)
        logging.info('-- learning_rate = %s', learning_rate)
        logging.info('-- momentum = %s', momentum)
        logging.info('-- nesterov = %s', nesterov)

        self._prepare(**kwargs)
        self._compile()

        if valid is None:
            valid = train
        iteration = 0
        training = validation = None
        while True:
            if not iteration % self.validate_every:
                try:
                    validation = self.evaluate(valid)
                except KeyboardInterrupt:
                    logging.info('interrupted!')
                    break
                if self._test_patience(validation):
                    logging.info('patience elapsed!')
                    break
            try:
                training = self._step(train)
            except KeyboardInterrupt:
                logging.info('interrupted!')
                break
            iteration += 1
            self._log(training, iteration)
            yield training, validation
        self.set_params(self._best_params)

    def minimize(self, *args, **kwargs):
        '''Optimize our loss exhaustively.

        This method is a thin wrapper over the :func:`iteropt` method. It simply
        exhausts the iterative optimization process and returns the final
        monitor values.

        Returns
        -------
        train_monitors : dict
            A dictionary mapping monitor names to values, evaluated on the
            training dataset.
        valid_monitors : dict
            A dictionary containing monitor values evaluated on the validation
            dataset.
        '''
        monitors = None
        for monitors in self.iteropt(*args, **kwargs):
            pass
        return monitors

    def _step(self, dataset):
        '''Advance the state of the optimizer by one step.

        Parameters
        ----------
        dataset : :class:`Dataset <downhill.dataset.Dataset>`
            A dataset for optimizing the model.

        Returns
        -------
        train_monitors : dict
            A dictionary mapping monitor names to values.
        '''
        values = [self.f_step(*x) for x in dataset]
        return collections.OrderedDict(
            zip(self._monitor_names, np.mean(values, axis=0)))



# -*- coding: utf-8 -*-

'''This module defines gradient descent optimizers with adaptive learning rates.
'''

import climate
import numpy as np
import theano
import theano.tensor as TT

from theano.sandbox.rng_mrg import MRG_RandomStreams as RandomStreams

from .base import Optimizer
from .util import as_float, shared_like

logging = climate.get_logger(__name__)

__all__ = ['RProp', 'RMSProp', 'ADADELTA', 'ESGD', 'Adam']


class RProp(Optimizer):
    r'''Optimization algorithm using resilient backpropagation.

    Like many stochastic optimizers, the RProp method takes small steps in
    parameter space using local gradient information. Unlike :class:`SGD
    <downhill.first_order.SGD>`, however, in RProp, only the signs of the
    gradients are taken into account when making parameter updates. That is, the
    step size for each parameter is independent of the magnitude of the gradient
    for that parameter.

    To accomplish this, RProp maintains a separate learning rate for every
    parameter in the model, and adjusts this learning rate based on the
    consistency of the sign of the gradient over time. Whenever two consecutive
    gradients for a parameter have the same sign, the learning rate for that
    parameter increases, and whenever the signs disagree, the learning rate
    decreases. This has a similar effect to momentum-based stochastic gradient
    methods but effectively maintains parameter-specific learning rates.

    .. math::
        \begin{eqnarray*}
        && \mbox{if } \frac{\partial\mathcal{L}}{\partial p}_{t-1}
           \frac{\partial\mathcal{L}}{\partial p} > 0 \\
        && \qquad \Delta_t = \min (\eta_+\Delta_{t−1}, \Delta_+) \\
        && \mbox{if } \frac{\partial\mathcal{L}}{\partial p}_{t-1}
           \frac{\partial\mathcal{L}}{\partial p} < 0 \\
        && \qquad \Delta_t = \max (\eta_-\Delta_{t−1}, \Delta_-) \\
        && \qquad \frac{\partial\mathcal{L}}{\partial p} = 0 \\
        && p_{t+1} = p_t − \mbox{sgn}\left(
           \frac{\partial\mathcal{L}}{\partial p}\right) \Delta_t
        \end{eqnarray*}

    Here, :math:`s(\cdot)` is the sign function (i.e., returns -1 if its
    argument is negative and 1 otherwise), :math:`\eta_-` and :math:`\eta_+` are
    the amount to decrease (increase) the step size if the gradients disagree
    (agree) in sign, and :math:`\Delta_+` and :math:`\Delta_-` are the maximum
    and minimum step size.

    The implementation here is actually the "iRprop-" variant of RProp described
    in Algorithm 4 from [Igel00]_. This variant resets the running gradient
    estimates to zero in cases where the previous and current gradients have
    switched signs.

    References
    ----------

    .. [Riedmiller92] M. Riedmiller & H. Braun. (1992) "Rprop - A Fast Adaptive
       Learning Algorithm." In Proceedings of the International Symposium on
       Computer and Information Science VII.

    .. [Igel00] C. Igel & M. Hüsken. (2000) "Improving the Rprop Learning
       Algorithm." In Proceedings of the Second International Symposium on Neural
       Computation. http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.17.1332
    '''

    def _prepare(self,
                 rprop_increase=1.01,
                 rprop_decrease=0.99,
                 rprop_min_step=0,
                 rprop_max_step=100,
                 **kwargs):
        self.step_increase = as_float(rprop_increase)
        self.step_decrease = as_float(rprop_decrease)
        self.min_step = as_float(rprop_min_step)
        self.max_step = as_float(rprop_max_step)
        logging.info('-- rprop_increase = %s', rprop_increase)
        logging.info('-- rprop_decrease = %s', rprop_decrease)
        logging.info('-- rprop_min_step = %s', rprop_min_step)
        logging.info('-- rprop_max_step = %s', rprop_max_step)
        super(RProp, self)._prepare(**kwargs)

    def _get_updates_for(self, param, grad):
        grad_tm1 = shared_like(param, 'grad')
        step_tm1 = shared_like(param, 'step', self.learning_rate.eval())
        test = grad * grad_tm1
        diff = TT.lt(test, 0)
        steps = step_tm1 * (TT.eq(test, 0) +
                            TT.gt(test, 0) * self.step_increase +
                            diff * self.step_decrease)
        step = TT.minimum(self.max_step, TT.maximum(self.min_step, steps))
        grad = grad - diff * grad
        yield param, param - TT.sgn(grad) * step
        yield grad_tm1, grad
        yield step_tm1, step


class RMSProp(Optimizer):
    r'''RMSProp optimizes scalar losses using scaled gradient steps.

    The RMSProp method uses the same general strategy as all first-order
    stochastic gradient methods, in the sense that these methods make small
    parameter adjustments iteratively using local derivative information.

    The difference here is that as gradients are computed during each parameter
    update, an exponentially-weighted moving average (EWMA) of gradient
    magnitudes is maintained as well. At each update, the EWMA is used to
    compute the root-mean-square (RMS) gradient value that's been seen in the
    recent past. The actual gradient is normalized by this RMS scaling factor
    before being applied to update the parameters.

    .. math::
        \begin{eqnarray*}
        f_{t+1} &=& \gamma a_t + (1 - \gamma) \frac{\partial\mathcal{L}}{\partial p} \\
        g_{t+1} &=& \gamma g_t + (1 - \gamma) \left(
           \frac{\partial\mathcal{L}}{\partial p}\right)^2 \\
        p_{t+1} &=& p_t - \frac{\alpha}{\sqrt{g_{t+1} - f_{t+1}^2 + \epsilon}}
           \frac{\partial\mathcal{L}}{\partial p}
        \end{eqnarray*}

    Like :class:`RProp`, this learning method effectively maintains a sort of
    parameter-specific momentum value, but this method takes into account both
    the sign and the magnitude of the gradient for each parameter.

    In this algorithm, RMS values are regularized (made less extreme) by
    :math:`\epsilon`, which is specified using the ``rms_regularizer`` keyword
    argument.

    The weight parameter :math:`\gamma` for the EWMA window is computed from the
    ``rms_halflife`` keyword argument, such that the actual EWMA weight varies
    inversely with the halflife :math:`h`: :math:`\gamma = e^{\frac{-\ln
    2}{h}}`.

    The implementation here is taken from [Graves13]_, equations (38)--(45).
    Graves' implementation in particular seems to have introduced the
    :math:`f_t` terms into the RMS computation; these terms appear to act as a
    sort of momentum for the RMS values.

    References
    ----------

    .. [Graves13] A. Graves. (2013) "Generating Sequences With Recurrent Neural
       Networks." http://arxiv.org/abs/1308.0850
    '''

    def _prepare(self, rms_halflife=14, rms_regularizer=1e-8, **kwargs):
        self.ewma = as_float(np.exp(-np.log(2) / rms_halflife))
        self.epsilon = as_float(rms_regularizer)
        logging.info('-- rms_halflife = %s', rms_halflife)
        logging.info('-- rms_regularizer = %s', rms_regularizer)
        super(RMSProp, self)._prepare(**kwargs)

    def _get_updates_for(self, param, grad):
        g1_tm1 = shared_like(param, 'g1_ewma')
        g2_tm1 = shared_like(param, 'g2_ewma')
        g1_t = self.ewma * g1_tm1 + (1 - self.ewma) * grad
        g2_t = self.ewma * g2_tm1 + (1 - self.ewma) * grad * grad
        rms = TT.sqrt(g2_t - g1_t * g1_t + self.epsilon)
        yield g1_tm1, g1_t
        yield g2_tm1, g2_t
        yield param, param - self.learning_rate * grad / rms


class ADADELTA(RMSProp):
    r'''ADADELTA optimizes scalar losses scaled stochastic gradient steps.

    The ADADELTA method uses the same general strategy as all first-order
    stochastic gradient methods, in the sense that these methods make small
    parameter adjustments iteratively using local derivative information.

    The difference with ADADELTA is that as gradients are computed during each
    parameter update, an exponentially-weighted weighted moving average (EWMA)
    gradient value, as well as an EWMA of recent parameter steps, are maintained
    as well. The actual gradient is normalized by the ratio of the
    root-mean-square (RMS) parameter step size to the RMS gradient magnitude.

    .. math::
        \begin{eqnarray*}
        g_{t+1} &=& \gamma g_t + (1 - \gamma) \left(
           \frac{\partial\mathcal{L}}{\partial p}\right)^2 \\
        v_{t+1} &=& \frac{\sqrt{x_t + \epsilon}}{\sqrt{g_{t+1} + \epsilon}}
           \frac{\partial\mathcal{L}}{\partial p} \\
        x_{t+1} &=& \gamma x_t + (1 - \gamma) v_{t+1}^2 \\
        p_{t+1} &=& p_t - v_{t+1}
        \end{eqnarray*}

    Like :class:`RProp` and the :class:`RMSProp`--:class:`ESGD` family, this
    learning method effectively maintains a sort of parameter-specific momentum
    value. The primary difference between this method and :class:`RMSProp` is
    that ADADELTA additionally incorporates a sliding window of RMS parameter
    step sizes, (somewhat) obviating the need for a learning rate parameter.

    In this implementation, the RMS values are regularized (made less extreme)
    by :math:`\epsilon`, which is specified using the ``rms_regularizer``
    parameter.

    The weight parameter :math:`\gamma` for the EWMA window is computed from the
    ``rms_halflife`` keyword argument, such that the actual EWMA weight varies
    inversely with the halflife :math:`h`: :math:`\gamma = e^{\frac{-\ln
    2}{h}}`.

    References
    ----------

    .. [Zeiler12] M. Zeiler. (2012) "ADADELTA: An adaptive learning rate method."
       http://arxiv.org/abs/1212.5701
    '''

    def _get_updates_for(self, param, grad):
        x2_tm1 = shared_like(param, 'x2_ewma')
        g2_tm1 = shared_like(param, 'g2_ewma')
        g2_t = self.ewma * g2_tm1 + (1 - self.ewma) * grad * grad
        delta = grad * TT.sqrt(x2_tm1 + self.epsilon) / TT.sqrt(g2_t + self.epsilon)
        x2_t = self.ewma * x2_tm1 + (1 - self.ewma) * delta * delta
        yield g2_tm1, g2_t
        yield x2_tm1, x2_t
        yield param, param - delta


class ESGD(RMSProp):
    r'''Equilibrated SGD computes a diagonal preconditioner for gradient descent.

    The ESGD method uses the same general strategy as all first-order
    stochastic gradient methods, in the sense that these methods make small
    parameter adjustments iteratively using local derivative information.

    The difference here is that as gradients are computed during each parameter
    update, an exponentially-weighted moving average (EWMA) of diagonal
    preconditioner estimates is maintained as well. At each update, the EWMA is
    used to compute the root-mean-square (RMS) diagonal preconditioner value
    that's been seen in the recent past. The actual gradient is normalized by
    this preconditioner before being applied to update the parameters.

    .. math::
        \begin{eqnarray*}
        r &\sim& \mathcal{N}(0, 1) \\
        Hr &=& \frac{\partial^2 \mathcal{L}}{\partial^2 p}r \\
        D_{t+1} &=& \gamma D_t + (1 - \gamma) (Hr)^2 \\
        p_{t+1} &=& p_t + - \frac{\alpha}{\sqrt{D_{t+1} + \epsilon}}
           \frac{\partial\mathcal{L}}{\partial p}
        \end{eqnarray*}

    Like :class:`Rprop` and the :class:`ADADELTA`--:class:`RMSProp` family, this
    learning method effectively maintains a sort of parameter-specific learning
    rate for each parameter in the loss.

    The primary difference between this method and :class:`RMSProp` is that ESGD
    treats the normalizing fraction explicitly as a preconditioner for the
    diaonal of the Hessian, and estimates this diagonal by drawing a vector of
    standard normal values at every training step.

    The primary difference between this implementation and the algorithm
    described in the paper (see below) is the use of an EWMA to decay the
    diagonal values over time, while in the paper the diagonal is divided by the
    training iteration. The EWMA halflife should be set to something reasonably
    large to ensure that this method emulates the method described in the
    original paper.

    In this implementation, :math:`\epsilon` regularizes the RMS values; it is
    is specified using the ``rms_regularizer`` parameter.

    The weight parameter :math:`\gamma` for the EWMA is computed from the
    ``rms_halflife`` keyword argument, such that the actual EWMA weight varies
    inversely with the halflife :math:`h`: :math:`\gamma = e^{\frac{-\ln
    2}{h}}`.

    References
    ----------

    .. [Dauphin14] Y. Dauphin, H. de Vries, J. Chung & Y. Bengio. (2014)
       "RMSProp and equilibrated adaptive learning rates for non-convex
       optimization." http://arxiv.org/abs/1502.04390
    '''

    def __init__(self, *args, **kwargs):
        self.rng = RandomStreams()
        super(ESGD, self).__init__(*args, **kwargs)

    def _get_updates_for(self, param, grad):
        D_tm1 = shared_like(param, 'D_ewma')
        Hv = TT.Rop(grad, param, self.rng.normal(param.shape))
        D_t = self.ewma * D_tm1 + (1 - self.ewma) * Hv * Hv
        den = TT.sqrt(D_t) + self.epsilon
        yield D_tm1, D_t
        yield param, param - grad * self.learning_rate / den


class Adam(RMSProp):
    r'''Adam optimizes using per-parameter learning rates.

    The ESGD method uses the same general strategy as all first-order
    stochastic gradient methods, in the sense that these methods make small
    parameter adjustments iteratively using local derivative information.

    The difference here is that as gradients are computed during each parameter
    update, exponentially-weighted moving averages (EWMAs) of (1) the first
    moment of the recent gradient values and (2) the second moment of recent
    gradient values are maintained as well. At each update, the step taken is
    proportional to the ratio of the first moment to the second moment.

    .. math::
        \begin{eqnarray*}
        \beta_1^t &=& \beta_1 \lambda^{t}
        f_{t+1} &=& \beta_1^t f_t + (1 - \beta_1^t)
           \frac{\partial\mathcal{L}}{\partial\theta} \\
        g_{t+1} &=& \beta_2 g_t + (1 - \beta_2)
           \left(\frac{\partial\mathcal{L}}{\partial\theta}\right)^2 \\
        \theta_{t+1} &=& \theta_t -
           \frac{f_{t+1} / (1 - \beta_1^t)}{\sqrt{g_{t+1} / (1 - \beta_2)} + \epsilon}
        \end{eqnarray*}

    Like all adaptive optimization algorithms, this optimizer effectively
    maintains a sort of parameter-specific momentum value. It shares with
    :class:`RMSProp` and :class:`ADADELTA` the idea of using an EWMA to track
    recent quantities related to the stochastic gradient during optimization.
    But the Adam method is unique in that it incorporates an explicit
    computation to remove the bias from these estimates.

    In this implementation, :math:`\epsilon` regularizes the RMS values and is
    given using the ``rms_regularizer`` keyword argument. The weight parameters
    :math:`\beta_1` and :math:`\beta_2` for the first and second EWMA windows
    are computed from the ``beta1_halflife`` and ``beta2_halflife`` keyword
    arguments, respectively, such that the actual EWMA weight varies inversely
    with the halflife :math:`h`: :math:`\gamma = e^{\frac{-\ln 2}{h}}`. The
    decay :math:`\lambda` for the :math:`\beta_1` EWMA is provided by the
    ``beta1_decay`` keyword argument.

    The implementation here is taken from Algorithm 1 of [Kin15]_.

    References
    ----------

    .. [Kin15] D. Kingma & J. Ba. (ICLR 2015) "Adam: A Method for
       Stochastic Optimization." http://arxiv.org/abs/1412.6980
    '''

    def _prepare(self,
                 beta1_decay=1 - 1e-6,
                 beta1_halflife=7,
                 beta2_halflife=69,
                 **kwargs):
        self.beta1_decay = as_float(beta1_decay)
        self.beta1 = as_float(np.exp(-np.log(2) / kwargs.get('beta1_halflife', 7)))
        self.beta2 = as_float(np.exp(-np.log(2) / kwargs.get('beta2_halflife', 69)))
        super(Adam, self)._prepare(**kwargs)

    def _get_updates_for(self, param, grad):
        t_tm1 = theano.shared(np.cast['float32'](0), 't')
        g1_tm1 = shared_like(param, 'g1_ewma')
        g2_tm1 = shared_like(param, 'g2_ewma')
        beta1 = self.beta1 * self.beta1_decay ** t_tm1
        g1_t = beta1 * g1_tm1 + (1 - beta1) * grad
        g2_t = self.beta2 * g2_tm1 + (1 - self.beta2) * grad * grad
        num = g1_t / (1 - beta1)
        den = TT.sqrt(g2_t / (1 - self.beta2))
        yield t_tm1, t_tm1 + 1
        yield g1_tm1, g1_t
        yield g2_tm1, g2_t
        yield param, param - self.learning_rate * num / (den + self.epsilon)



# -*- coding: utf-8 -*-

'''This module defines first-order gradient descent optimizers.'''

import climate

from .base import Optimizer

logging = climate.get_logger(__name__)

__all__ = ['SGD', 'NAG']


class SGD(Optimizer):
    r'''Optimize using stochastic gradient descent with momentum.

    A stochastic gradient trainer with momentum :math:`\mu` and learning rate
    :math:`\alpha` updates parameter :math:`\theta` at step :math:`t` by
    blending the current "velocity" :math:`v` with the current gradient
    :math:`\frac{\partial\mathcal{L}}{\partial\theta}`:

    .. math::
        \begin{eqnarray*}
        v_{t+1} &=& \mu v_t - \alpha \frac{\partial\mathcal{L}}{\partial\theta} \\
        \theta_{t+1} &=& \theta_t + v_{t+1}
        \end{eqnarray*}

    Without momentum (i.e., when :math:`\mu = 0`), these updates reduce to
    :math:`\theta_{t+1} = \theta_t - \alpha \frac{\partial\mathcal{L}}{\partial\theta}`,
    which just takes steps downhill according to the the local gradient.

    Adding the momentum term permits the algorithm to incorporate information
    from previous steps as well, which in practice is thought to have the effect
    of incorporating some information about second-order derivatives of the loss
    surface.
    '''

    def _get_updates_for(self, param, grad):
        yield param, param - self.learning_rate * grad


class NAG(SGD):
    r'''Optimize using Nesterov's Accelerated Gradient (NAG).

    The basic difference between NAG and "classical" momentum in SGD
    optimization approaches is that NAG computes the gradients at the position
    in parameter space where "classical" momentum would put us at the *next*
    step. In classical :class:`SGD` with momentum :math:`\mu` and learning rate
    :math:`\alpha`, updates to parameter :math:`p` at step :math:`t` are
    computed by blending the current "velocity" :math:`v` with the current
    gradient :math:`\frac{\partial\mathcal{L}}{\partial p}`:

    .. math::
        \begin{eqnarray*}
        v_{t+1} &=& \mu v_t - \alpha \frac{\partial\mathcal{L}}{\partial p} \\
        p_{t+1} &=& p_t + v_{t+1}
        \end{eqnarray*}

    In contrast, NAG adjusts the update by blending the current "velocity" with
    the gradient at the next step---that is, the gradient is computed at the
    point where the velocity would have taken us:

    .. math::
        \begin{eqnarray*}
        v_{t+1} &=& \mu v_t - \alpha \left.
           \frac{\partial\mathcal{L}}{\partial p}\right|_{p_t + \mu v_t} \\
        p_{t+1} &=& p_t + v_{t+1}
        \end{eqnarray*}

    Again, the difference here is that the gradient is computed at the place in
    parameter space where we would have stepped using the classical technique,
    in the absence of a new gradient.

    In theory, this helps correct for oversteps during learning: If momentum
    would lead us to overshoot, then the gradient at that overshot place will
    point backwards, toward where we came from. See [1]_ for details on this idea.

    References
    ----------
    .. [1] I. Sutskever, J. Martens, G. Dahl, & G. Hinton. (ICML 2013) "On the
       importance of initialization and momentum in deep learning."
       http://jmlr.csail.mit.edu/proceedings/papers/v28/sutskever13.pdf
    '''

    def _prepare(self, **kwargs):
        super(NAG, self)._prepare(**kwargs)
        self.nesterov = True



from .adaptive import *
from .base import build, Optimizer
from .dataset import Dataset
from .first_order import *


def minimize(loss, params, inputs, train, valid=None,
             algo='rmsprop', updates=(), monitors=(),
             monitor_gradients=False, batch_size=32,
             train_batches=None, valid_batches=None,
             **kwargs):
    '''Minimize a loss function with respect to some symbolic parameters.

    Parameters
    ----------
    loss : Theano expression
        Loss function to minimize. This must be a scalar-valued expression.
    params : list of Theano variables
        Symbolic variables to adjust to minimize the loss.
    inputs : list of Theano variables
        Symbolic variables required to compute the loss.
    train : :class:`Dataset`, ndarray, or callable
        Dataset to use for computing gradient updates.
    valid : :class:`Dataset`, ndarray, or callable, optional
        Dataset to use for validating the minimization process. The training
        dataset is used if this is not provided.
    algo : str, optional
        Name of the minimization algorithm to use. Must be one of the strings
        that can be passed to :func:`build`. Defaults to ``'rmsprop'``.
    updates : list of update pairs, optional
        A list of pairs providing updates for the internal of the loss
        computation. Normally this is empty, but it can be provided if the loss,
        for example, requires an update to an internal random number generator.
    monitors : dict or sequence of (str, Theano expression) tuples, optional
        Additional values to monitor during optimization. These must be provided
        as either a sequence of (name, expression) tuples, or as a dictionary
        mapping string names to Theano expressions.
    monitor_gradients : bool, optional
        If True, add monitors to log the norms of the parameter gradients during
        optimization. Defaults to False.
    batch_size : int, optional
        Size of batches provided by datasets. Defaults to 32.
    train_batches : int, optional
        Number of batches of training data to iterate over during one pass of
        optimization. Defaults to None, which uses the entire training dataset.
    valid_batches : int, optional
        Number of batches of validation data to iterate over during one pass of
        validation. Defaults to None, which uses the entire validation dataset.

    Additional keyword arguments are passed to the optimizer instance.

    Returns
    -------
    train_monitors : dict
        A dictionary mapping monitor names to monitor values. This dictionary
        will always contain the ``'loss'`` key, giving the value of the loss
        evaluated on the training dataset.
    valid_monitors : dict
        A dictionary mapping monitor names to monitor values, evaluated on the
        validation dataset. This dictionary will always contain the ``'loss'``
        key, giving the value of the loss function. Because validation is not
        always computed after every optimization update, these monitor values
        may be "stale"; however, they will always contain the most recently
        computed values.
    '''
    if not isinstance(train, Dataset):
        train = Dataset(
            train,
            name='train',
            batch_size=batch_size,
            iteration_size=train_batches,
        )
    if valid is not None and not isinstance(valid, Dataset):
        valid = Dataset(
            valid,
            name='valid',
            batch_size=batch_size,
            iteration_size=valid_batches,
        )
    return build(
        algo,
        loss=loss,
        params=params,
        inputs=inputs,
        updates=updates,
        monitors=monitors,
        monitor_gradients=monitor_gradients,
    ).minimize(train, valid, **kwargs)



# -*- coding: utf-8 -*-

'''This module defines miscellaneous optimizers.'''

import climate
import numpy as np
import scipy.optimize
import theano.tensor as TT

from .base import Optimizer

logging = climate.get_logger(__name__)


class Scipy(Optimizer):
    '''General trainer for neural nets using ``scipy``.

    This class serves as a wrapper for the optimization algorithms provided in
    `scipy.optimize.minimize`_. The following algorithms are available in this
    trainer:

    - ``bfgs``
    - ``cg``
    - ``dogleg``
    - ``newton-cg``
    - ``trust-ncg``

    In general, these methods require two types of computations in order to
    minimize a cost function: evaluating the cost function for a specific
    setting of model parameters, and computing the gradient of the cost function
    for a specific setting of model parameters. Both of these computations are
    implemented by the ``downhill`` package and may, if you have a GPU, involve
    computing values on the GPU.

    However, all of the optimization steps that might be performed once these
    two types of values are computed will not be handled on the GPU, since
    ``scipy`` is not capable of using the GPU. This might or might not influence
    the absolute time required to optimize a model, depending on the ratio of
    time spent computing cost and gradient values to the time spent computing
    parameter updates.

    For more information about these optimization methods, please see the `Scipy
    documentation`_.

    .. _scipy.optimize.minimize: http://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
    .. _Scipy documentation: http://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
    '''

    METHODS = ('bfgs', 'cg', 'dogleg', 'newton-cg', 'trust-ncg')

    def __init__(self, method, *args, **kwargs):
        super(Scipy, self).__init__(*args, **kwargs)

        self.method = method

    def compile(self):
        super(Scipy, self).compile()
        logging.info('compiling gradient function')
        self.f_grad = theano.function(self.inputs, TT.grad(self.loss, self.params))

    def flat_to_arrays(self, x):
        '''Convert a parameter vector to a sequence of parameter arrays.

        Parameters
        ----------
        flat : ndarray
            A one-dimensional numpy array containing flattened parameter values
            for all parameters in our model.

        Returns
        -------
        arrays : sequence of ndarray
            Values of the parameters in our model.
        '''
        x = x.astype(self._dtype)
        return [x[o:o+n].reshape(s) for s, o, n in
                zip(self._shapes, self._starts, self._counts)]

    def arrays_to_flat(self, arrays):
        '''Convert a sequence of parameter arrays to a vector.

        Parameters
        ----------
        arrays : sequence of ndarray
            Values of the parameters in our model.

        Returns
        -------
        flat : ndarray
            A one-dimensional numpy array containing flattened parameter values
            for all parameters in our model.
        '''
        x = np.zeros((sum(self._counts), ), self._dtype)
        for arr, o, n in zip(arrays, self._starts, self._counts):
            x[o:o+n] = arr.ravel()
        return x

    def function_at(self, x, dataset):
        '''Compute the value of the loss function at given parameter values.

        Parameters
        ----------
        x : ndarray
            An array of parameter values to set our model at.
        dataset : :class:`Dataset <downhill.dataset.Dataset>`
            A set of data over which to compute our loss function.

        Returns
        -------
        loss : float
            Scalar value of the loss function, evaluated at the given parameter
            settings, using the given dataset.
        '''
        self.set_params(self.flat_to_arrays(x))
        return self.evaluate(dataset)['loss']

    def gradient_at(self, x, dataset):
        '''Compute the gradients of the loss function at given parameter values.

        Parameters
        ----------
        x : ndarray
            An array of parameter values to set our model at.
        dataset : :class:`Dataset <downhill.dataset.Dataset>`
            A set of data over which to compute our gradients.

        Returns
        -------
        gradients : ndarray
            A vector of gradient values, of the same dimensions as `x`.
        '''
        self.set_params(self.flat_to_arrays(x))
        grads = [[] for _ in range(len(self.params))]
        for x in dataset:
            for i, g in enumerate(self.f_grad(*x)):
                grads[i].append(np.asarray(g))
        return self.arrays_to_flat([np.mean(g, axis=0) for g in grads])

    def step(self, dataset):
        '''Advance the state of the model by one training step.

        Parameters
        ----------
        dataset : :class:`Dataset <downhill.dataset.Dataset>`
            A dataset for training the model.

        Returns
        -------
        training : dict
            A dictionary mapping monitor names to values.
        '''
        res = scipy.optimize.minimize(
            fun=self.function_at,
            jac=self.gradient_at,
            x0=self.arrays_to_flat(self._best_params),
            args=(dataset, ),
            method=self.method,
            options=dict(maxiter=self.validate_every),
        )
        self.set_params(self.flat_to_arrays(res.x))
        return self.evaluate(dataset)



import os
import setuptools

setuptools.setup(
    name='downhill',
    version='0.1.0',
    packages=setuptools.find_packages(),
    author='Leif Johnson',
    author_email='leif@lmjohns3.com',
    description='Stochastic optimization routines for Theano',
    long_description=open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'README.rst')).read(),
    license='MIT',
    url='http://github.com/lmjohns3/downhill',
    keywords=('adadelta '
              'adam '
              'esgd '
              'gradient-descent '
              'nesterov '
              'optimization '
              'rmsprop '
              'sgd '
              'theano '
              ),
    install_requires=['theano', 'climate'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering',
        ],
    )
